import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
from collections import OrderedDict
from torch.nn import init
import numpy as np

def get_upsample_filter(size):
    '''
    Make a 2D bilinear kernel suitable for upsampling
    '''
    factor = (size + 1)//2
    if size % 2 == 1:
        center = factor - 1
    else:
        center = factor - 0.5
    og = np.ogrid[:size, :size]
    filter = (1 - abs(og[0] - center) / factor) * \
            (1 - abs(og[1] - center) / factor)
    return torch.form_numpy(filter).float()


class FCN(nn.Module):

    def __init__(self, n_classes=3):
        super().__init__()
        self.mode_name = 'FCN'
        self.n_classes = n_classes

        self.conv_block1 = nn.Sequential(
                            nn.Conv2d(1, 64, 3, padding=100),
                            nn.ReLU(inplace=True),
                            nn.Conv2d(64,64,3, padding=1),
                            nn.ReLU(inplace=True),
                            nn.MaxPool2d(2,stride=2,ceil_mode=True),
                            )
        self.conv_block2 = nn.Sequential(
                            nn.Conv2d(64, 128, 3, padding=1),
                            nn.ReLU(inplace=True),
                            nn.Conv2d(128,128,3, padding=1),
                            nn.ReLU(inplace=True),
                            nn.MaxPool2d(2,stride=2,ceil_mode=True),
                            )
        self.conv_block3 = nn.Sequential(
                            nn.Conv2d(128, 256, 3, padding=1),
                            nn.ReLU(inplace=True),
                            nn.Conv2d(256,256,3, padding=1),
                            nn.ReLU(inplace=True),
                            nn.MaxPool2d(2,stride=2,ceil_mode=True),
                            )
        self.conv_block4 = nn.Sequential(
                            nn.Conv2d(256, 512, 3, padding=1),
                            nn.ReLU(inplace=True),
                            nn.Conv2d(512,512,3, padding=1),
                            nn.ReLU(inplace=True),
                            nn.MaxPool2d(2,stride=2,ceil_mode=True),
                            )
        self.conv_block5 = nn.Sequential(
                            nn.Conv2d(512, 512, 3, padding=1),
                            nn.ReLU(inplace=True),
                            nn.Conv2d(512,512,3, padding=1),
                            nn.ReLU(inplace=True),
                            nn.MaxPool2d(2,stride=2,ceil_mode=True),
                            )
        self.classifier = nn.Sequential(
                            nn.Conv2d(512, 4096, 7),
                            nn.ReLU(inplace=True),
                            nn.Dropout2d(),
                            nn.Conv2d(4096,4096,1),
                            nn.ReLU(inplace=True),
                            nn.Dropout2d(),
                            nn.Conv2d(4096, self.n_classes,1),
                            )

        self.score_pool4 = nn.Conv2d(512, self.n_classes, 1)
        self.score_pool3 = nn.Conv2d(256, self.n_classes, 1)

        self.upscore = nn.ConvTranspose2d(self.n_classes, self.n_classes, 16, stride=8)
        self.upscore4 = nn.ConvTranspose2d(self.n_classes, self.n_classes, 4, stride=2)
        self.upscore5 = nn.ConvTranspose2d(self.n_classes, self.n_classes, 4, stride=2)


    def forward(self, x):
        conv1 = self.conv_block1(x)
        conv2 = self.conv_block2(conv1)
        conv3 = self.conv_block3(conv2)
        conv4 = self.conv_block4(conv3)
        conv5 = self.conv_block5(conv4)
        score5 = self.classifier(conv5)

        upscore5 = self.upscore5(score5)
        score4 = self.score_pool4(conv4)
        score4 = score4[:,:,5:5+upscore5.size()[2], 5:5+upscore5.size()[3]].contiguous()
        score4 += upscore5

        upscore4 = self.upscore4(score4)
        score3 = self.score_pool3(conv3)
        score3 = score3[:,:,9:9+upscore4.size()[2], 9:9+upscore4.size()[3]].contiguous()
        score3 += upscore4

        out = self.upscore(score3)
        out = out[:,:, 31:31+x.size()[2], 31:31+x.size()[3]].contiguous()
        #print(out.shape)
        return out


    def init(self):
        vgg16 = models.vgg16(pretrained=True)
        self.init_vgg16_params(vgg16)
'''
device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
model = FCN().to(device)

l = 0.01
X = torch.ones(3,1,240,240)
X = X.to(device)
#print(X.shape)
Y = 10*torch.zeros(3,1,240,240)
Y = Y.to(device)
#print(Y.shape)
optimizer = torch.optim.Adam(model.parameters(), lr=l)
y= model(X)
print(y.shape)



def rewrite(x):
    y = (x > 0.5).astype('uint8')
    return y

def conv3x3(in_channels, out_channels, stride=1,
            padding=1, bias=True, groups=1):
    return nn.Conv2d(
        in_channels,
        out_channels,
        kernel_size=3,
        stride=stride,
        padding=padding,
        bias=bias,
        groups=groups)

def upconv2x2(in_channels, out_channels, mode='transpose'):
    if mode == 'transpose':
        return nn.ConvTranspose2d(
            in_channels,
            out_channels,
            kernel_size=2,
            stride=2)
    else:
        return nn.Sequential(
            nn.Upsample(mode='bilinear', scale_factor=2),
            conv1x1(in_channels, out_channels))

def conv1x1(in_channels, out_channels, groups=1):
    return nn.Conv2d(
        in_channels,
        out_channels,
        kernel_size=1,
        groups=groups,
        stride=1)


class DownConv(nn.Module):

    def __init__(self, in_channels, out_channels, pooling=True):
        super(DownConv, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.pooling = pooling

        self.conv1 = conv3x3(self.in_channels, self.out_channels)
        self.conv2 = conv3x3(self.out_channels, self.out_channels)

        if self.pooling:
            self.pool = nn.MaxPool2d(kernel_size=2, stride=2)

    def forward(self, x):
        x = F.relu(self.conv1(x))
        x = F.relu(self.conv2(x))
        before_pool = x     #for connecting directly
        if self.pooling:
            x = self.pool(x)
        return x, before_pool


class UpConv(nn.Module):

    def __init__(self, in_channels, out_channels,
                 merge_mode='add', up_mode='transpose'):
        super(UpConv, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.merge_mode = merge_mode
        self.up_mode = up_mode

        self.upconv = upconv2x2(self.in_channels, self.out_channels,
            mode=self.up_mode)

        if self.merge_mode == 'concat':
            self.conv1 = conv3x3(
                2*self.out_channels, self.out_channels)
        else:
            self.conv1 = conv3x3(self.out_channels, self.out_channels)
        self.conv2 = conv3x3(self.out_channels, self.out_channels)


    def forward(self, from_down, from_up):
        from_up = self.upconv(from_up)
        if self.merge_mode == 'concat':
            x = torch.cat((from_up, from_down), 1)
        else:
            x = from_up + from_down
        x = F.relu(self.conv1(x))
        x = F.relu(self.conv2(x))
        return x


class FCN(nn.Module):

    def __init__(self, num_classes, in_channels=1, depth=5,
                 start_filts=64, up_mode='transpose',
                 merge_mode='add'):

        super(FCN, self).__init__()

        if up_mode in ('transpose', 'upsample'):
            self.up_mode = up_mode
        else:
            raise ValueError("\"{}\" is not a valid mode for "
                             "upsampling. Only \"transpose\" and "
                             "\"upsample\" are allowed.".format(up_mode))

        if merge_mode in ('concat', 'add'):
            self.merge_mode = merge_mode
        else:
            raise ValueError("\"{}\" is not a valid mode for"
                             "merging up and down paths. "
                             "Only \"concat\" and "
                             "\"add\" are allowed.".format(up_mode))

        # NOTE: up_mode 'upsample' is incompatible with merge_mode 'add'
        if self.up_mode == 'upsample' and self.merge_mode == 'add':
            raise ValueError("up_mode \"upsample\" is incompatible "
                             "with merge_mode \"add\" at the moment "
                             "because it doesn't make sense to use "
                             "nearest neighbour to reduce "
                             "depth channels (by half).")

        self.num_classes = num_classes
        self.in_channels = in_channels
        self.start_filts = start_filts
        self.depth = depth

        self.down_convs = []
        self.up_convs = []

        for i in range(depth):
            ins = self.in_channels if i == 0 else outs
            outs = self.start_filts*(2**i)
            pooling = True if i < depth-1 else False

            down_conv = DownConv(ins, outs, pooling=pooling)
            self.down_convs.append(down_conv)

        for i in range(depth-1):
            ins = outs
            outs = ins // 2
            up_conv = UpConv(ins, outs, up_mode=up_mode,
                merge_mode=merge_mode)
            self.up_convs.append(up_conv)

        self.conv_final = conv1x1(outs, self.num_classes)
        self.conv_final_re = conv1x1(outs,self.in_channels)

        self.down_convs = nn.ModuleList(self.down_convs)
        self.up_convs = nn.ModuleList(self.up_convs)

        self.reset_params()

    @staticmethod
    def weight_init(m):
        if isinstance(m, nn.Conv2d):
            nn.init.kaiming_uniform_(m.weight)
            nn.init.constant_(m.bias, 0)


    def reset_params(self):
        for i, m in enumerate(self.modules()):
            self.weight_init(m)


    def forward(self, x):
        encoder_outs = []

        for i, module in enumerate(self.down_convs):
            x, before_pool = module(x)
            encoder_outs.append(before_pool)

        seg = x
        for i, module in enumerate(self.up_convs):
            before_pool = encoder_outs[-(i+2)]
            seg = module(before_pool, seg)

        seg = self.conv_final(seg)
        seg = torch.sigmoid(seg)
        return seg
'''
